import * as i0 from '@angular/core';
import { Directive, Inject, Input, HostListener, EventEmitter, Component, ChangeDetectionStrategy, Output, ContentChildren, NgModule } from '@angular/core';
import * as i2 from '@angular/common';
import { CommonModule } from '@angular/common';
import * as i1 from '@angular/platform-browser';

class CellSize {
    constructor(rowCount, columnCount) {
        this._height = 0;
        this._width = 0;
        this._height = rowCount ? 100 / rowCount : 0;
        this._width = columnCount ? 100 / columnCount : 0;
    }
    ;
    get height() {
        return this._height;
    }
    get width() {
        return this._width;
    }
}

class Grid {
    constructor(rows, columns) {
        this._widgets = [];
        this._rows = 3;
        this._columns = 3;
        if (+rows) {
            this._rows = +rows;
        }
        if (+columns) {
            this._columns = +columns;
        }
        this._cellSize = new CellSize(this._rows, this._columns);
    }
    get widgets() {
        return this._widgets;
    }
    get rows() {
        return this._rows;
    }
    get columns() {
        return this._columns;
    }
    get cellSize() {
        return this._cellSize;
    }
    add(widget) {
        this._widgets.push(widget);
    }
    remove(widget) {
        const widgetIndex = this._widgets.indexOf(widget);
        if (widgetIndex > -1) {
            this._widgets.splice(widgetIndex, 1);
        }
    }
    removeAll() {
        this._widgets.splice(0);
    }
    hasWidget(widget) {
        const widgetIndex = this._widgets.indexOf(widget);
        return widgetIndex > -1;
    }
    resize(rows, columns) {
        columns = +columns || 0;
        rows = +rows || 0;
        if (columns > 0 && rows > 0 && columns !== this._columns || rows !== this._rows) {
            this._columns = columns;
            this._rows = rows;
            this._cellSize = new CellSize(this._rows, this._columns);
        }
    }
}

class Rectangle {
    constructor(obj) {
        this.top = 0;
        this.left = 0;
        this.width = 0;
        this.height = 0;
        if (obj) {
            this.top = +obj.top || 0;
            this.left = +obj.left || 0;
            this.width = +obj.width || 0;
            this.height = +obj.height || 0;
        }
    }
    get bottom() {
        return this.top + this.height - 1;
    }
    get right() {
        return this.left + this.width - 1;
    }
    getSurfaceArea() {
        return this.height * this.width;
    }
}

class WidgetConfig {
    constructor(rect) {
        this._position = new Rectangle();
        this.id = this.generateUID();
        if (rect) {
            this.position = rect;
        }
    }
    get position() {
        return this._position;
    }
    set position(gridArea) {
        this._position.top = +gridArea.top ? +gridArea.top : 0;
        this._position.left = +gridArea.left ? +gridArea.left : 0;
        this._position.width = +gridArea.width ? +gridArea.width : 0;
        this._position.height = +gridArea.height ? +gridArea.height : 0;
    }
    generateUID() {
        return 'ngxDashboardWidget-' + ++WidgetConfig.widgetCount;
    }
}
WidgetConfig.widgetCount = 0;

class Utils {
    static isNumber(val) {
        return typeof val === 'number';
    }
    static isDefined(val) {
        return val !== undefined;
    }
    static isObject(val) {
        return typeof val === 'object';
    }
}
var RESIZE_DIRECTIONS;
(function (RESIZE_DIRECTIONS) {
    RESIZE_DIRECTIONS["topLeft"] = "NW";
    RESIZE_DIRECTIONS["top"] = "N";
    RESIZE_DIRECTIONS["topRight"] = "NE";
    RESIZE_DIRECTIONS["right"] = "E";
    RESIZE_DIRECTIONS["bottomRight"] = "SE";
    RESIZE_DIRECTIONS["bottom"] = "S";
    RESIZE_DIRECTIONS["bottomLeft"] = "SW";
    RESIZE_DIRECTIONS["left"] = "W";
})(RESIZE_DIRECTIONS || (RESIZE_DIRECTIONS = {}));
const ALL_DIRECTIONS = [
    RESIZE_DIRECTIONS.bottom,
    RESIZE_DIRECTIONS.left,
    RESIZE_DIRECTIONS.right,
    RESIZE_DIRECTIONS.top,
    RESIZE_DIRECTIONS.bottomLeft,
    RESIZE_DIRECTIONS.bottomRight,
    RESIZE_DIRECTIONS.topLeft,
    RESIZE_DIRECTIONS.topRight
];

class Cell {
    constructor(top, left) {
        this._top = top;
        this._left = left;
    }
    get top() {
        return this._top;
    }
    get left() {
        return this._left;
    }
}

class PathIterator {
    constructor(start, end) {
        this._currPos = null;
        this._nextPos = null;
        if (!start) {
            console.error('Start not present for Path Iterator');
            return;
        }
        if (!end) {
            console.error('End not present for Path Iterator');
            return;
        }
        this._start = start;
        this._heightDelta = end.top - start.top;
        this._widthDelta = end.left - start.left;
        this._steps = Math.max(Math.abs(this._heightDelta), Math.abs(this._widthDelta));
        this._currStep = 0;
        this._currPos = null;
        this._nextPos = new Cell(start.top, start.left);
    }
    next() {
        this._currPos = this._nextPos;
        if (this._currStep < this._steps) {
            this._currStep++;
            const currTopDelta = Math.round((this._currStep / this._steps) * this._heightDelta);
            const currLeftDelta = Math.round((this._currStep / this._steps) * this._widthDelta);
            this._nextPos = new Cell(this._start.top + currTopDelta, this._start.left + currLeftDelta);
        }
        else {
            this._nextPos = null;
        }
        return this._currPos;
    }
    hasNext() {
        return this._nextPos !== null;
    }
}

class NgxWidgetMoverDirective {
    constructor(el, renderer, gridCmp, widgetCmp) {
        this.el = el;
        this.renderer = renderer;
        this.gridCmp = gridCmp;
        this.widgetCmp = widgetCmp;
        this.enableDrag = null;
        this._onMoveListener = this.onMove.bind(this);
        this._onUpListener = this.onUp.bind(this);
        this.ngxWidgetMover = false;
    }
    onDown(event) {
        if (event.button === 0) {
            event.preventDefault();
            this.renderer.addClass(this.widgetCmp.getEl().nativeElement, 'wg-moving');
            const widgetContainer = this.widgetCmp.getEl().nativeElement;
            this.startPosition = this.gridCmp.getWidgetPosition(this.widgetCmp);
            this.startRender = {
                top: widgetContainer.offsetTop,
                left: widgetContainer.offsetLeft,
                height: widgetContainer.clientHeight,
                width: widgetContainer.clientWidth
            }; // pixel values
            const eventOffsetX = event.offsetX || event.layerX;
            const eventOffsetY = event.offsetY || event.layerY;
            this.desiredPosition = { top: this.startRender.top, left: this.startRender.left };
            this.moverOffset = new Rectangle({
                top: eventOffsetY + this.el.nativeElement.offsetTop || 0,
                left: eventOffsetX + this.el.nativeElement.offsetLeft || 0
            });
            this.gridPositions = this.gridCmp.getGridRectangle();
            this.cellHeight = (this.gridCmp.grid.cellSize.height / 100) * this.gridPositions.height;
            this.cellWidth = (this.gridCmp.grid.cellSize.width / 100) * this.gridPositions.width;
            this.enableDrag = this.widgetCmp.getConfig().id;
            this.renderer.setStyle(this.widgetCmp.getEl().nativeElement, 'z-index', this.ngxWidgetMover ? 0 : 100);
            if (typeof PointerEvent !== 'undefined') {
                window.addEventListener('pointermove', this._onMoveListener);
                window.addEventListener('pointerup', this._onUpListener);
            }
            else {
                window.addEventListener('mousemove', this._onMoveListener);
                window.addEventListener('mouseup', this._onUpListener);
            }
        }
    }
    onMove(event) {
        if (this.enableDrag === this.widgetCmp.getConfig().id) {
            event.preventDefault();
            const eventClientX = event.clientX;
            const eventClientY = event.clientY;
            const startRender = this.startRender;
            const gridDimensions = this.gridPositions;
            const desiredPosition = this.desiredPosition;
            // normalize the drag position
            const dragPositionX = Math.round(eventClientX) - gridDimensions.left, dragPositionY = Math.round(eventClientY) - gridDimensions.top;
            desiredPosition.top = Math.min(Math.max(dragPositionY - this.moverOffset.top, 0), gridDimensions.height - startRender.height - 1);
            desiredPosition.left = Math.min(Math.max(dragPositionX - this.moverOffset.left, 0), gridDimensions.width - startRender.width - 1);
            const currentFinalPos = this.determineFinalPos(this.startPosition, desiredPosition, this.startRender, this.cellHeight, this.cellWidth);
            this.gridCmp.highlightArea(currentFinalPos);
            this.renderer.setStyle(this.widgetCmp.getEl().nativeElement, 'top', desiredPosition.top + 'px');
            this.renderer.setStyle(this.widgetCmp.getEl().nativeElement, 'left', desiredPosition.left + 'px');
        }
    }
    onUp(event) {
        if (this.enableDrag === this.widgetCmp.getConfig().id) {
            event.preventDefault();
            const eventClientX = event.clientX;
            const eventClientY = event.clientY;
            const startRender = this.startRender;
            const gridDimensions = this.gridPositions;
            const desiredPosition = this.desiredPosition;
            // normalize the drag position
            const dragPositionX = Math.round(eventClientX) - gridDimensions.left, dragPositionY = Math.round(eventClientY) - gridDimensions.top;
            desiredPosition.top = Math.min(Math.max(dragPositionY - this.moverOffset.top, 0), gridDimensions.height - startRender.height - 1);
            desiredPosition.left = Math.min(Math.max(dragPositionX - this.moverOffset.left, 0), gridDimensions.width - startRender.width - 1);
            const anchorTop = this.getAnchor(Math.max(dragPositionY, 0), this.cellHeight, 1);
            const anchorLeft = this.getAnchor(Math.max(dragPositionX, 0), this.cellWidth, 1);
            const dropPosition = this.gridCmp.rasterizeCoords(anchorLeft, anchorTop);
            const obstructingWidgetId = this.gridCmp.areaObstructor(dropPosition);
            let finalPos;
            if (obstructingWidgetId && this.ngxWidgetMover) {
                const obstructingWidgetCmp = this.gridCmp.getWidgetById(obstructingWidgetId);
                const obstructingWidgetPosition = this.gridCmp.getWidgetPositionByWidgetId(obstructingWidgetId);
                const draggedWidgetPosition = this.widgetCmp.position;
                this.widgetCmp.position = obstructingWidgetPosition;
                this.gridCmp.updateWidget(this.widgetCmp, true);
                obstructingWidgetCmp.position = draggedWidgetPosition;
                this.gridCmp.updateWidget(obstructingWidgetCmp, true);
            }
            else {
                finalPos = this.determineFinalPos(this.startPosition, desiredPosition, this.startRender, this.cellHeight, this.cellWidth);
                this.widgetCmp.position = finalPos;
                this.gridCmp.updateWidget(this.widgetCmp, false);
            }
            this.gridCmp.resetHighlights();
            this.renderer.removeClass(this.widgetCmp.getEl().nativeElement, 'wg-moving');
            this.renderer.removeStyle(this.widgetCmp.getEl().nativeElement, 'z-index');
            this.enableDrag = null;
        }
        if (typeof PointerEvent !== 'undefined') {
            window.removeEventListener('pointermove', this._onMoveListener);
            window.removeEventListener('pointerup', this._onUpListener);
        }
        else {
            window.removeEventListener('mousemove', this._onMoveListener);
            window.removeEventListener('mouseup', this._onUpListener);
        }
    }
    getAnchor(val, cellWOrH, marginFactor = 2) {
        return (val % cellWOrH) > (cellWOrH / marginFactor) ? val + Math.floor(cellWOrH) : val;
    }
    determineFinalPos(startPos, desiredPos, startRender, cellHt, cellWd) {
        if (startRender.top === desiredPos.top && startRender.left === desiredPos.left) {
            return startPos;
        }
        const anchorTop = this.getAnchor(desiredPos.top, cellHt);
        const anchorLeft = this.getAnchor(desiredPos.left, cellWd);
        const movedDown = anchorTop >= startRender.top;
        const movedRight = anchorLeft >= startRender.left;
        const desiredFinalPosition = this.gridCmp.rasterizeCoords(anchorLeft, anchorTop);
        const path = new PathIterator(desiredFinalPosition, startPos);
        while (path && path.hasNext()) {
            const currPos = path.next();
            const targetArea = new Rectangle({
                top: currPos.top,
                left: currPos.left,
                height: startPos.height,
                width: startPos.width
            });
            const options = {
                excludedArea: startPos,
                fromBottom: movedDown,
                fromRight: movedRight
            };
            // If widget swap is enabled and area is obstructed, show original position
            if (this.ngxWidgetMover && this.gridCmp.isAreaObstructed(targetArea, options)) {
                return new Rectangle(startPos);
            }
            if (!this.gridCmp.isAreaObstructed(targetArea, options)) {
                // try to get closer to the desired position by leaving the original path
                const height = targetArea.height;
                const width = targetArea.width;
                if (desiredFinalPosition.top < targetArea.top) {
                    while (desiredFinalPosition.top <= (targetArea.top - 1)) {
                        const checkRect = new Rectangle({ top: targetArea.top - 1, left: targetArea.left, height, width });
                        const isRectVacant = !this.gridCmp.isAreaObstructed(checkRect, options);
                        if (isRectVacant) {
                            targetArea.top--;
                        }
                        else {
                            break;
                        }
                    }
                }
                else if (desiredFinalPosition.top > targetArea.top) {
                    while (desiredFinalPosition.top >= (targetArea.top + 1)) {
                        const checkRect = new Rectangle({ top: targetArea.top + 1, left: targetArea.left, height, width });
                        const isRectVacant = !this.gridCmp.isAreaObstructed(checkRect, options);
                        if (isRectVacant) {
                            targetArea.top++;
                        }
                        else {
                            break;
                        }
                    }
                }
                if (desiredFinalPosition.left < targetArea.left) {
                    while (desiredFinalPosition.left <= (targetArea.left - 1)) {
                        const checkRect = new Rectangle({ top: targetArea.top, left: targetArea.left - 1, height, width });
                        const isRectVacant = !this.gridCmp.isAreaObstructed(checkRect, options);
                        if (isRectVacant) {
                            targetArea.left--;
                        }
                        else {
                            break;
                        }
                    }
                }
                else if (desiredFinalPosition.left > targetArea.left) {
                    while (desiredFinalPosition.left >= (targetArea.left + 1)) {
                        const checkRect = new Rectangle({ top: targetArea.top, left: targetArea.left + 1, height, width });
                        const isRectVacant = !this.gridCmp.isAreaObstructed(checkRect, options);
                        if (isRectVacant) {
                            targetArea.left++;
                        }
                        else {
                            break;
                        }
                    }
                }
                return new Rectangle(targetArea);
            }
        }
        return new Rectangle(startPos);
    }
}
NgxWidgetMoverDirective.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "15.2.9", ngImport: i0, type: NgxWidgetMoverDirective, deps: [{ token: i0.ElementRef }, { token: i0.Renderer2 }, { token: 'NgxWidgetGridComponent' }, { token: 'NgxWidgetComponent' }], target: i0.ɵɵFactoryTarget.Directive });
NgxWidgetMoverDirective.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "14.0.0", version: "15.2.9", type: NgxWidgetMoverDirective, selector: "[ngxWidgetMover]", inputs: { ngxWidgetMover: "ngxWidgetMover" }, host: { listeners: { "mousedown": "onDown($event)", "pointerdown": "onDown($event)" } }, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "15.2.9", ngImport: i0, type: NgxWidgetMoverDirective, decorators: [{
            type: Directive,
            args: [{
                    selector: '[ngxWidgetMover]'
                }]
        }], ctorParameters: function () { return [{ type: i0.ElementRef }, { type: i0.Renderer2 }, { type: undefined, decorators: [{
                    type: Inject,
                    args: ['NgxWidgetGridComponent']
                }] }, { type: undefined, decorators: [{
                    type: Inject,
                    args: ['NgxWidgetComponent']
                }] }]; }, propDecorators: { ngxWidgetMover: [{
                type: Input
            }], onDown: [{
                type: HostListener,
                args: ['mousedown', ['$event']]
            }, {
                type: HostListener,
                args: ['pointerdown', ['$event']]
            }] } });

const MIN_HEIGHT = 42;
const MIN_WIDTH = 42;
class NgxWidgetResizerDirective {
    constructor(el, renderer, gridCmp, widgetCmp) {
        this.el = el;
        this.renderer = renderer;
        this.gridCmp = gridCmp;
        this.widgetCmp = widgetCmp;
        this.moveUpAllowed = false;
        this.moveDownAllowed = false;
        this.moveLeftAllowed = false;
        this.moveRightAllowed = false;
        this.enableDrag = null;
        this._onMoveListener = this.onMove.bind(this);
        this._onUpListener = this.onUp.bind(this);
        this.parentContainer = this.el.nativeElement.parentElement;
    }
    get resizeDirection() {
        return this._resizeDirection;
    }
    set ngxWidgetResizer(dir) {
        this._resizeDirection = dir;
        this.moveUpAllowed = false;
        this.moveDownAllowed = false;
        this.moveLeftAllowed = false;
        this.moveRightAllowed = false;
        switch (dir) {
            case RESIZE_DIRECTIONS.top:
                this.moveUpAllowed = true;
                break;
            case RESIZE_DIRECTIONS.left:
                this.moveLeftAllowed = true;
                break;
            case RESIZE_DIRECTIONS.bottom:
                this.moveDownAllowed = true;
                break;
            case RESIZE_DIRECTIONS.right:
                this.moveRightAllowed = true;
                break;
            case RESIZE_DIRECTIONS.topLeft:
                this.moveUpAllowed = true;
                this.moveLeftAllowed = true;
                break;
            case RESIZE_DIRECTIONS.topRight:
                this.moveUpAllowed = true;
                this.moveRightAllowed = true;
                break;
            case RESIZE_DIRECTIONS.bottomLeft:
                this.moveDownAllowed = true;
                this.moveLeftAllowed = true;
                break;
            case RESIZE_DIRECTIONS.bottomRight:
                this.moveDownAllowed = true;
                this.moveRightAllowed = true;
                break;
            default:
        }
    }
    onDown(event) {
        event.preventDefault();
        this.enableDrag = this.widgetCmp.getConfig().id;
        this.renderer.addClass(this.widgetCmp.getEl().nativeElement, 'wg-resizing');
        this.renderer.addClass(this.el.nativeElement, 'dragging');
        this.startPosition = this.gridCmp.getWidgetPosition(this.widgetCmp);
        this.startRender = {
            top: Math.ceil(this.widgetCmp.getEl().nativeElement.offsetTop),
            left: Math.ceil(this.widgetCmp.getEl().nativeElement.offsetLeft),
            height: Math.floor(this.parentContainer.offsetHeight),
            width: Math.floor(this.parentContainer.offsetWidth)
        }; // pixel values
        this.startRender.bottom = this.startRender.top + this.startRender.height;
        this.startRender.right = this.startRender.left + this.startRender.width;
        const eventOffsetX = event.offsetX || event.layerX;
        const eventOffsetY = event.offsetY || event.layerY;
        this.delta = { top: 0, right: 0, bottom: 0, left: 0 };
        this.draggerOffset = {
            top: eventOffsetY,
            left: eventOffsetX,
            bottom: eventOffsetY - this.el.nativeElement.offsetHeight,
            right: eventOffsetX - this.el.nativeElement.offsetWidth
        };
        this.gridPositions = this.gridCmp.getGridRectangle();
        if (typeof PointerEvent !== 'undefined') {
            window.addEventListener('pointermove', this._onMoveListener);
            window.addEventListener('pointerup', this._onUpListener);
        }
        else {
            window.addEventListener('mousemove', this._onMoveListener);
            window.addEventListener('mouseup', this._onUpListener);
        }
    }
    onMove(event) {
        if (this.enableDrag === this.widgetCmp.getConfig().id) {
            event.preventDefault();
            const eventClientX = event.clientX;
            const eventClientY = event.clientY;
            const gridDims = this.gridPositions;
            const startRender = this.startRender;
            // normalize the drag position
            const dragPositionX = Math.round(eventClientX) - gridDims.left;
            const dragPositionY = Math.round(eventClientY) - gridDims.top;
            const delta = this.delta;
            if (this.moveUpAllowed) {
                delta.top = Math.min(Math.max(dragPositionY - this.draggerOffset.top, 0), gridDims.height) - startRender.top;
                delta.top = Math.min(delta.top, startRender.height - MIN_HEIGHT);
            }
            else if (this.moveDownAllowed) {
                delta.bottom = startRender.bottom - Math.min(Math.max(dragPositionY - this.draggerOffset.bottom, 0), gridDims.height);
                delta.bottom = Math.min(delta.bottom, startRender.height - MIN_HEIGHT);
            }
            if (this.moveLeftAllowed) {
                delta.left = Math.min(Math.max(dragPositionX - this.draggerOffset.left, 0), gridDims.width) - startRender.left;
                delta.left = Math.min(delta.left, startRender.width - MIN_WIDTH);
            }
            else if (this.moveRightAllowed) {
                delta.right = startRender.right - Math.min(Math.max(dragPositionX - this.draggerOffset.right, 0), gridDims.width);
                delta.right = Math.min(delta.right, startRender.width - MIN_WIDTH);
            }
            const currentFinalPos = this.determineFinalPos();
            this.gridCmp.highlightArea(currentFinalPos);
            this.renderer.setStyle(this.parentContainer, 'top', this.delta.top + 'px');
            this.renderer.setStyle(this.parentContainer, 'left', this.delta.left + 'px');
            this.renderer.setStyle(this.parentContainer, 'bottom', this.delta.bottom + 'px');
            this.renderer.setStyle(this.parentContainer, 'right', this.delta.right + 'px');
        }
    }
    onUp(event) {
        if (this.enableDrag === this.widgetCmp.getConfig().id) {
            event.preventDefault();
            this.el.nativeElement.setAttribute('draggable', false);
            this.renderer.removeClass(this.el.nativeElement, 'dragging');
            this.renderer.removeClass(this.widgetCmp.getEl().nativeElement, 'wg-resizing');
            this.enableDrag = null;
            this.widgetCmp.position = this.determineFinalPos();
            this.gridCmp.updateWidget(this.widgetCmp, false);
            this.gridCmp.resetHighlights();
            // reset style
            this.renderer.removeClass(this.widgetCmp.getEl().nativeElement, 'wg-resizing');
            this.renderer.removeClass(this.el.nativeElement, 'dragging');
            this.renderer.setStyle(this.parentContainer, 'top', '');
            this.renderer.setStyle(this.parentContainer, 'left', '');
            this.renderer.setStyle(this.parentContainer, 'bottom', '');
            this.renderer.setStyle(this.parentContainer, 'right', '');
        }
        if (typeof PointerEvent !== 'undefined') {
            window.removeEventListener('pointermove', this._onMoveListener);
            window.removeEventListener('pointerup', this._onUpListener);
        }
        else {
            window.removeEventListener('mousemove', this._onMoveListener);
            window.removeEventListener('mouseup', this._onUpListener);
        }
    }
    findCollision(start, end, val, reverse = false) {
        let foundCollision = false;
        for (let i = start; i <= end; i++) {
            const checker = reverse ? this.gridCmp.isPointObstructed(i, val) : this.gridCmp.isPointObstructed(val, i);
            if (checker) {
                foundCollision = true;
                break;
            }
        }
        return foundCollision;
    }
    determineFinalPos() {
        const finalPos = new Rectangle();
        const startRender = this.startRender;
        const delta = this.delta;
        const requestedStartPoint = this.gridCmp.rasterizeCoords(startRender.left + delta.left + 1, startRender.top + delta.top + 1);
        const requestedEndPoint = this.gridCmp.rasterizeCoords(startRender.right - delta.right - 1, startRender.bottom - delta.bottom - 1);
        const requestedPos = {
            top: requestedStartPoint.top,
            left: requestedStartPoint.left,
            right: requestedEndPoint.left,
            bottom: requestedEndPoint.top
        };
        // determine a suitable final position (one that is not obstructed)
        let foundCollision;
        const start = Math.max(this.startPosition.left, requestedPos.left);
        const end = Math.min(this.startPosition.right, requestedPos.right);
        if (this.moveUpAllowed && requestedPos.top < this.startPosition.top) {
            finalPos.top = this.startPosition.top;
            while (finalPos.top > requestedPos.top) {
                // check whether adding another row above would cause any conflict
                foundCollision = this.findCollision(start, end, finalPos.top - 1);
                if (foundCollision) {
                    break;
                }
                finalPos.top--; // add row above
            }
        }
        else if (this.moveDownAllowed && requestedPos.bottom > this.startPosition.bottom) {
            finalPos.top = finalPos.top || requestedPos.top;
            finalPos.height = this.startPosition.bottom + 1 - this.startPosition.top;
            while (finalPos.bottom < requestedPos.bottom) {
                // check whether adding another row below would cause any conflict
                foundCollision = this.findCollision(start, end, finalPos.bottom + 1);
                if (foundCollision) {
                    break;
                }
                finalPos.height++; // add row below
            }
        }
        finalPos.top = finalPos.top || requestedPos.top;
        finalPos.height = finalPos.height || requestedPos.bottom + 1 - finalPos.top;
        if (this.moveLeftAllowed && requestedPos.left < this.startPosition.left) {
            finalPos.left = this.startPosition.left;
            while (finalPos.left > requestedPos.left) {
                // check whether adding another column would cause any conflict
                foundCollision = this.findCollision(finalPos.top, finalPos.bottom, finalPos.left - 1, true);
                if (foundCollision) {
                    break;
                }
                finalPos.left--; // add column
            }
        }
        else if (this.moveRightAllowed && requestedPos.right > this.startPosition.right) {
            finalPos.left = finalPos.left || requestedPos.left;
            finalPos.width = this.startPosition.right + 1 - this.startPosition.left;
            while (finalPos.right < requestedPos.right) {
                foundCollision = this.findCollision(finalPos.top, finalPos.bottom, finalPos.right + 1, true);
                if (foundCollision) {
                    break;
                }
                finalPos.width++;
            }
        }
        finalPos.left = finalPos.left || requestedPos.left;
        finalPos.width = finalPos.width || requestedPos.right + 1 - finalPos.left;
        return finalPos;
    }
}
NgxWidgetResizerDirective.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "15.2.9", ngImport: i0, type: NgxWidgetResizerDirective, deps: [{ token: i0.ElementRef }, { token: i0.Renderer2 }, { token: 'NgxWidgetGridComponent' }, { token: 'NgxWidgetComponent' }], target: i0.ɵɵFactoryTarget.Directive });
NgxWidgetResizerDirective.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "14.0.0", version: "15.2.9", type: NgxWidgetResizerDirective, selector: "[ngxWidgetResizer]", inputs: { ngxWidgetResizer: "ngxWidgetResizer" }, host: { listeners: { "pointerdown": "onDown($event)", "mousedown": "onDown($event)" } }, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "15.2.9", ngImport: i0, type: NgxWidgetResizerDirective, decorators: [{
            type: Directive,
            args: [{
                    selector: '[ngxWidgetResizer]'
                }]
        }], ctorParameters: function () { return [{ type: i0.ElementRef }, { type: i0.Renderer2 }, { type: undefined, decorators: [{
                    type: Inject,
                    args: ['NgxWidgetGridComponent']
                }] }, { type: undefined, decorators: [{
                    type: Inject,
                    args: ['NgxWidgetComponent']
                }] }]; }, propDecorators: { ngxWidgetResizer: [{
                type: Input
            }], onDown: [{
                type: HostListener,
                args: ['pointerdown', ['$event']]
            }, {
                type: HostListener,
                args: ['mousedown', ['$event']]
            }] } });

class NgxWidgetComponent {
    constructor(elRef) {
        this.elRef = elRef;
        this.swapOnMove = false;
        this.positionChange = new EventEmitter();
        this.movable = false;
        this.allDirections = RESIZE_DIRECTIONS;
        this.isTopResizable = false;
        this.isRightResizable = false;
        this.isBottomResizable = false;
        this.isLeftResizable = false;
        this.isTopRightResizable = false;
        this.isTopLeftResizable = false;
        this.isBottomRightResizable = false;
        this.isBottomLeftResizable = false;
        this._resizable = false;
        this._resizeDirections = ALL_DIRECTIONS;
        this.widgetConfig = new WidgetConfig(this.position);
    }
    get position() {
        return this._position;
    }
    set position(position) {
        this._position = position;
        this.widgetConfig.position = position;
        this.positionChange.emit(position);
    }
    get resizable() {
        return this._resizable;
    }
    set resizable(resizable) {
        this._resizable = resizable;
        if (resizable) {
            this.setResizeDirections();
        }
    }
    get resizeDirections() {
        return this._resizeDirections;
    }
    set resizeDirections(dirs) {
        this._resizeDirections = dirs.filter((dir) => {
            return ALL_DIRECTIONS.indexOf(dir.toUpperCase()) !== -1;
        });
        this.setResizeDirections();
    }
    setResizeDirections() {
        this.isTopResizable = false;
        this.isRightResizable = false;
        this.isBottomResizable = false;
        this.isLeftResizable = false;
        this.isTopRightResizable = false;
        this.isTopLeftResizable = false;
        this.isBottomRightResizable = false;
        this.isBottomLeftResizable = false;
        this._resizeDirections.forEach((dir) => {
            switch (dir) {
                case RESIZE_DIRECTIONS.top:
                    this.isTopResizable = true;
                    break;
                case RESIZE_DIRECTIONS.left:
                    this.isLeftResizable = true;
                    break;
                case RESIZE_DIRECTIONS.bottom:
                    this.isBottomResizable = true;
                    break;
                case RESIZE_DIRECTIONS.right:
                    this.isRightResizable = true;
                    break;
                case RESIZE_DIRECTIONS.topLeft:
                    this.isTopLeftResizable = true;
                    break;
                case RESIZE_DIRECTIONS.topRight:
                    this.isTopRightResizable = true;
                    break;
                case RESIZE_DIRECTIONS.bottomLeft:
                    this.isBottomLeftResizable = true;
                    break;
                case RESIZE_DIRECTIONS.bottomRight:
                    this.isBottomRightResizable = true;
                    break;
                default:
            }
        });
    }
    getConfig() {
        return this.widgetConfig;
    }
    getEl() {
        return this.elRef;
    }
}
NgxWidgetComponent.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "15.2.9", ngImport: i0, type: NgxWidgetComponent, deps: [{ token: i0.ElementRef }], target: i0.ɵɵFactoryTarget.Component });
NgxWidgetComponent.ɵcmp = i0.ɵɵngDeclareComponent({ minVersion: "14.0.0", version: "15.2.9", type: NgxWidgetComponent, selector: "ngx-widget", inputs: { swapOnMove: "swapOnMove", movable: "movable", position: "position", resizable: "resizable", resizeDirections: "resizeDirections" }, outputs: { positionChange: "positionChange" }, ngImport: i0, template: "<div class=\"wg-widget-content\">\n  <ng-content></ng-content>\n</div>\n<div [ngxWidgetMover]=\"swapOnMove\"\n     *ngIf=\"movable\"\n     class=\"wg-widget-edit-move\">\n</div>\n<div *ngIf=\"resizable\"\n     class=\"wg-widget-edit-resize\">\n  <div class=\"wg-resize wg-resize-axis wg-resize-n\"\n       [ngxWidgetResizer]=\"allDirections.top\"\n       [hidden]=\"!isTopResizable\"></div>\n  <div class=\"wg-resize wg-resize-axis wg-resize-e\"\n       [ngxWidgetResizer]=\"allDirections.right\"\n       [hidden]=\"!isRightResizable\"></div>\n  <div class=\"wg-resize wg-resize-axis wg-resize-s\"\n       [ngxWidgetResizer]=\"allDirections.bottom\"\n       [hidden]=\"!isBottomResizable\"></div>\n  <div class=\"wg-resize wg-resize-axis wg-resize-w\"\n       [ngxWidgetResizer]=\"allDirections.left\"\n       [hidden]=\"!isLeftResizable\"></div>\n  <div class=\"wg-resize wg-resize-diag wg-resize-nw\"\n       [ngxWidgetResizer]=\"allDirections.topLeft\"\n       [hidden]=\"!isTopLeftResizable\"></div>\n  <div class=\"wg-resize wg-resize-diag wg-resize-ne\"\n       [ngxWidgetResizer]=\"allDirections.topRight\"\n       [hidden]=\"!isTopRightResizable\"></div>\n  <div class=\"wg-resize wg-resize-diag wg-resize-se\"\n       [ngxWidgetResizer]=\"allDirections.bottomRight\"\n       [hidden]=\"!isBottomRightResizable\"></div>\n  <div class=\"wg-resize wg-resize-diag wg-resize-sw\"\n       [ngxWidgetResizer]=\"allDirections.bottomLeft\"\n       [hidden]=\"!isBottomLeftResizable\"></div>\n</div>\n\n", styles: [":host{position:absolute}:host .wg-widget-content{position:absolute;overflow:hidden;height:100%;width:100%}:host.wg-moving .wg-widget-content,:host.wg-resizing .wg-widget-content{opacity:.2;z-index:18}:host.wg-moving .wg-widget-edit-move{cursor:move}.wg-widget-edit-move{position:absolute;inset:6px;z-index:20;touch-action:none;cursor:move}.wg-widget-edit-resize{position:absolute;touch-action:none;inset:0;z-index:15}.wg-resize{position:absolute}.wg-resize-diag{border-color:#0071bcbf;border-style:solid;width:0;height:0;z-index:1}.wg-resize-diag:hover{border-color:#008ae5bf}.wg-resize-diag.dragging{border-color:#00aeefbf}.wg-resize-axis{background-color:#0071bc1f}.wg-resize-axis.dragging{background-color:#00aeef3d}.wg-resize-nw{top:0;left:0;cursor:nw-resize;border-width:.5rem .5rem 0 0;border-right-color:transparent!important}.wg-resize-ne{top:0;right:0;cursor:ne-resize;border-width:.5rem 0 0 .5rem;border-left-color:transparent!important}.wg-resize-se{bottom:0;right:0;cursor:se-resize;border-width:0 0 .5rem .5rem;border-left-color:transparent!important}.wg-resize-sw{bottom:0;left:0;cursor:sw-resize;border-width:0 .5rem .5rem 0;border-right-color:transparent!important}.wg-resize-n{top:0;right:0;left:0;height:.25rem;cursor:n-resize}.wg-resize-e{top:0;right:0;bottom:0;width:.25rem;cursor:e-resize}.wg-resize-s{right:0;bottom:0;left:0;height:.25rem;cursor:s-resize}.wg-resize-w{top:0;bottom:0;left:0;width:.25rem;cursor:w-resize}[hidden]{display:none}\n"], dependencies: [{ kind: "directive", type: i2.NgIf, selector: "[ngIf]", inputs: ["ngIf", "ngIfThen", "ngIfElse"] }, { kind: "directive", type: NgxWidgetMoverDirective, selector: "[ngxWidgetMover]", inputs: ["ngxWidgetMover"] }, { kind: "directive", type: NgxWidgetResizerDirective, selector: "[ngxWidgetResizer]", inputs: ["ngxWidgetResizer"] }], changeDetection: i0.ChangeDetectionStrategy.OnPush });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "15.2.9", ngImport: i0, type: NgxWidgetComponent, decorators: [{
            type: Component,
            args: [{ selector: 'ngx-widget', changeDetection: ChangeDetectionStrategy.OnPush, template: "<div class=\"wg-widget-content\">\n  <ng-content></ng-content>\n</div>\n<div [ngxWidgetMover]=\"swapOnMove\"\n     *ngIf=\"movable\"\n     class=\"wg-widget-edit-move\">\n</div>\n<div *ngIf=\"resizable\"\n     class=\"wg-widget-edit-resize\">\n  <div class=\"wg-resize wg-resize-axis wg-resize-n\"\n       [ngxWidgetResizer]=\"allDirections.top\"\n       [hidden]=\"!isTopResizable\"></div>\n  <div class=\"wg-resize wg-resize-axis wg-resize-e\"\n       [ngxWidgetResizer]=\"allDirections.right\"\n       [hidden]=\"!isRightResizable\"></div>\n  <div class=\"wg-resize wg-resize-axis wg-resize-s\"\n       [ngxWidgetResizer]=\"allDirections.bottom\"\n       [hidden]=\"!isBottomResizable\"></div>\n  <div class=\"wg-resize wg-resize-axis wg-resize-w\"\n       [ngxWidgetResizer]=\"allDirections.left\"\n       [hidden]=\"!isLeftResizable\"></div>\n  <div class=\"wg-resize wg-resize-diag wg-resize-nw\"\n       [ngxWidgetResizer]=\"allDirections.topLeft\"\n       [hidden]=\"!isTopLeftResizable\"></div>\n  <div class=\"wg-resize wg-resize-diag wg-resize-ne\"\n       [ngxWidgetResizer]=\"allDirections.topRight\"\n       [hidden]=\"!isTopRightResizable\"></div>\n  <div class=\"wg-resize wg-resize-diag wg-resize-se\"\n       [ngxWidgetResizer]=\"allDirections.bottomRight\"\n       [hidden]=\"!isBottomRightResizable\"></div>\n  <div class=\"wg-resize wg-resize-diag wg-resize-sw\"\n       [ngxWidgetResizer]=\"allDirections.bottomLeft\"\n       [hidden]=\"!isBottomLeftResizable\"></div>\n</div>\n\n", styles: [":host{position:absolute}:host .wg-widget-content{position:absolute;overflow:hidden;height:100%;width:100%}:host.wg-moving .wg-widget-content,:host.wg-resizing .wg-widget-content{opacity:.2;z-index:18}:host.wg-moving .wg-widget-edit-move{cursor:move}.wg-widget-edit-move{position:absolute;inset:6px;z-index:20;touch-action:none;cursor:move}.wg-widget-edit-resize{position:absolute;touch-action:none;inset:0;z-index:15}.wg-resize{position:absolute}.wg-resize-diag{border-color:#0071bcbf;border-style:solid;width:0;height:0;z-index:1}.wg-resize-diag:hover{border-color:#008ae5bf}.wg-resize-diag.dragging{border-color:#00aeefbf}.wg-resize-axis{background-color:#0071bc1f}.wg-resize-axis.dragging{background-color:#00aeef3d}.wg-resize-nw{top:0;left:0;cursor:nw-resize;border-width:.5rem .5rem 0 0;border-right-color:transparent!important}.wg-resize-ne{top:0;right:0;cursor:ne-resize;border-width:.5rem 0 0 .5rem;border-left-color:transparent!important}.wg-resize-se{bottom:0;right:0;cursor:se-resize;border-width:0 0 .5rem .5rem;border-left-color:transparent!important}.wg-resize-sw{bottom:0;left:0;cursor:sw-resize;border-width:0 .5rem .5rem 0;border-right-color:transparent!important}.wg-resize-n{top:0;right:0;left:0;height:.25rem;cursor:n-resize}.wg-resize-e{top:0;right:0;bottom:0;width:.25rem;cursor:e-resize}.wg-resize-s{right:0;bottom:0;left:0;height:.25rem;cursor:s-resize}.wg-resize-w{top:0;bottom:0;left:0;width:.25rem;cursor:w-resize}[hidden]{display:none}\n"] }]
        }], ctorParameters: function () { return [{ type: i0.ElementRef }]; }, propDecorators: { swapOnMove: [{
                type: Input
            }], positionChange: [{
                type: Output
            }], movable: [{
                type: Input
            }], position: [{
                type: Input
            }], resizable: [{
                type: Input
            }], resizeDirections: [{
                type: Input
            }] } });

class GridRenderer {
    constructor(grid) {
        this.positions = {};
        this.obstructions = [];
        this.grid = grid || new Grid();
    }
    get grid() {
        return this._grid;
    }
    set grid(grid) {
        this._grid = grid;
        this.positions = {};
        this.cachedNextPosition = undefined;
        this.obstructions = [];
        for (let i = 0; i < grid.rows * grid.columns; i++) {
            this.obstructions[i] = null;
        }
    }
    rasterizeCoords(left, top, gridWidth, gridHeight) {
        left = Math.min(Math.max(left, 0), gridWidth - 1);
        top = Math.min(Math.max(top, 0), gridHeight - 1);
        const x = Math.floor(top / gridHeight * this.grid.rows) + 1;
        const y = Math.floor(left / gridWidth * this.grid.columns) + 1;
        return new Cell(x, y);
    }
    getWidgetIdAt(i, j) {
        for (const widgetId in this.positions) {
            if (this.positions.hasOwnProperty(widgetId)) {
                const position = this.positions[widgetId];
                if (position.top <= i && i <= (position.top + position.height - 1) &&
                    position.left <= j && j <= (position.left + position.width - 1)) {
                    return widgetId;
                }
            }
        }
        return null;
    }
    getWidgetPosition(widgetId) {
        return this.positions[widgetId];
    }
    setWidgetPosition(widgetId, newPosition, swappingPositions) {
        const currPosition = this.positions[widgetId];
        if (currPosition && !swappingPositions) {
            this.setObstructionValues(currPosition, null);
        }
        newPosition = new Rectangle({
            top: Utils.isNumber(newPosition.top) ? newPosition.top : currPosition.top,
            left: Utils.isNumber(newPosition.left) ? newPosition.left : currPosition.left,
            width: Utils.isNumber(newPosition.width) ? newPosition.width : currPosition.width,
            height: Utils.isNumber(newPosition.height) ? newPosition.height : currPosition.height
        });
        this.positions[widgetId] = newPosition;
        this.setObstructionValues(this.positions[widgetId], widgetId);
        this.cachedNextPosition = undefined;
    }
    hasSpaceLeft() {
        for (const obstruction of this.obstructions) {
            if (obstruction === null) {
                return true;
            }
        }
        return false;
    }
    getNextPosition() {
        if (this.cachedNextPosition !== undefined) {
            return this.cachedNextPosition;
        }
        if (!this.hasSpaceLeft()) {
            return null;
        }
        const maxPosition = this.findLargestEmptyArea();
        this.cachedNextPosition = maxPosition;
        return maxPosition;
    }
    isObstructed(i, j, excludedArea) {
        // obstructed if (i, j) exceeds the grid's regular non-expanding boundaries
        if (i < 1 || j < 1 || j > this.grid.columns || i > this.grid.rows) {
            return true;
        }
        const bottom = excludedArea && excludedArea.top + excludedArea.height - 1;
        const right = excludedArea && excludedArea.left + excludedArea.width - 1;
        // pass if (i, j) is within the excluded area, if any
        if (excludedArea &&
            excludedArea.top <= i &&
            i <= bottom &&
            excludedArea.left <= j &&
            j <= right) {
            return false;
        }
        return this._isObstructed(i, j);
    }
    _isObstructed(i, j) {
        return this.obstructions[(i - 1) * this.grid.columns + (j - 1)] !== null;
    }
    isAreaObstructed(area, options) {
        if (!area) {
            return false;
        }
        options = Utils.isObject(options) ? options : {};
        const top = area.top;
        const left = area.left;
        const bottom = area.bottom || area.top + area.height - 1;
        const right = area.right || area.left + area.width - 1;
        if (!Utils.isNumber(top) || !Utils.isNumber(left) || !Utils.isNumber(bottom) || !Utils.isNumber(right)) {
            return false;
        }
        const verticalStart = options.fromBottom ? bottom : top;
        const verticalStep = options.fromBottom ? -1 : 1;
        const verticalEnd = (options.fromBottom ? top : bottom) + verticalStep;
        const horizontalStart = options.fromRight ? right : left;
        const horizontalStep = options.fromRight ? -1 : 1;
        const horizontalEnd = (options.fromRight ? left : right) + horizontalStep;
        for (let i = verticalStart; i !== verticalEnd; i += verticalStep) {
            for (let j = horizontalStart; j !== horizontalEnd; j += horizontalStep) {
                if (this.isObstructed(i, j, options.excludedArea)) {
                    return true;
                }
            }
        }
        return false;
    }
    getStyle(widgetId) {
        const render = this.positions[widgetId];
        if (!render) {
            return { display: 'none' };
        }
        return {
            top: ((render.top - 1) * this.grid.cellSize.height).toString() + '%',
            height: (render.height * this.grid.cellSize.height).toString() + '%',
            left: ((render.left - 1) * this.grid.cellSize.width).toString() + '%',
            width: (render.width * this.grid.cellSize.width).toString() + '%'
        };
    }
    setObstructionValues(area, value) {
        for (let i = area.top - 1; i < area.top + area.height - 1; i++) {
            for (let j = area.left - 1; j < area.left + area.width - 1; j++) {
                this.obstructions[i * this.grid.columns + j] = value;
            }
        }
    }
    findLargestEmptyArea() {
        let maxArea = null;
        let currMaxArea = null;
        let maxSurfaceArea = 0;
        let currMaxSurfaceArea = 0;
        for (let i = 1; i <= this.grid.rows; i++) {
            for (let j = 1; j <= this.grid.columns; j++) {
                if (this._isObstructed(i, j)) {
                    continue;
                }
                const currAreaLimit = (this.grid.rows - i + 1) * (this.grid.columns - j + 1);
                if (currAreaLimit < maxSurfaceArea) {
                    break;
                }
                currMaxArea = this._findLargestEmptyAreaFrom(new Cell(i, j));
                currMaxSurfaceArea = currMaxArea.getSurfaceArea();
                if (currMaxSurfaceArea > maxSurfaceArea) {
                    maxSurfaceArea = currMaxSurfaceArea;
                    maxArea = currMaxArea;
                }
            }
        }
        return maxArea;
    }
    _findLargestEmptyAreaFrom(start) {
        if (!Utils.isDefined(this.grid) || !Utils.isNumber(this.grid.columns) || !Utils.isNumber(this.grid.rows)) {
            return null;
        }
        let maxArea = null, maxSurfaceArea = 0, endColumn = this.grid.columns;
        for (let i = start.top; i <= this.grid.rows; i++) {
            for (let j = start.left; j <= endColumn; j++) {
                if (this._isObstructed(i, j)) {
                    endColumn = j - 1;
                    continue;
                }
                const currHeight = (i - start.top + 1);
                const currWidth = (j - start.left + 1);
                const currSurfaceArea = currHeight * currWidth;
                if (currSurfaceArea > maxSurfaceArea) {
                    maxSurfaceArea = currSurfaceArea;
                    maxArea = new Rectangle({
                        top: start.top,
                        left: start.left,
                        width: currWidth,
                        height: currHeight
                    });
                }
            }
        }
        return maxArea;
    }
    render(grid, emitWidgetPositionUpdated) {
        this.grid = grid;
        const widgets = this.grid && this.grid.widgets ? this.grid.widgets : [];
        const unpositionedWidgets = [];
        widgets.forEach((widget) => {
            const position = widget.position;
            if (position.width * position.height === 0 ||
                this.isAreaObstructed(position)) {
                unpositionedWidgets.push(widget);
            }
            else {
                this.setWidgetPosition(widget.id, position, false);
            }
        });
        unpositionedWidgets.forEach((widget) => {
            const nextPosition = this.getNextPosition();
            if (nextPosition !== null) {
                widget.position = nextPosition;
                this.setWidgetPosition(widget.id, nextPosition, false);
            }
            else {
                widget.position = new Rectangle();
                this.setWidgetPosition(widget.id, new Rectangle(), false);
            }
            if (emitWidgetPositionUpdated) {
                emitWidgetPositionUpdated(widget);
            }
        });
    }
}

class NgxGridOverlayComponent {
    constructor(sanitizer) {
        this.sanitizer = sanitizer;
        this.activeHighlight = null;
        this.gridRows = [];
        this.gridCols = [];
        this._showGrid = false;
    }
    get renderer() {
        return this._renderer;
    }
    set renderer(renderer) {
        this._renderer = renderer;
        if (Utils.isDefined(renderer)) {
            this.updateGridLines(renderer, this.showGrid);
        }
    }
    set rows(rows) {
        this.updateGridLines(this.renderer, this.showGrid);
    }
    set cols(rows) {
        this.updateGridLines(this.renderer, this.showGrid);
    }
    get highlight() {
        return this._highlight;
    }
    set highlight(highlight) {
        this._highlight = highlight;
        this.clearHighlight();
        if (highlight) {
            this.highlightArea(highlight, this.renderer);
        }
    }
    get showGrid() {
        return this._showGrid;
    }
    set showGrid(showGrid) {
        this._showGrid = showGrid;
        this.updateGridLines(this.renderer, showGrid);
    }
    updateGridLines(renderer, showGrid) {
        this.clearGridLines();
        if (showGrid) {
            this.showGridLines(renderer);
        }
    }
    clearHighlight() {
        this.activeHighlight = null;
    }
    clearGridLines() {
        this.gridRows.splice(0);
        this.gridCols.splice(0);
    }
    highlightArea(area, renderer) {
        const cellSize = renderer.grid.cellSize;
        const cellHeight = cellSize.height;
        const cellWidth = cellSize.width;
        this.activeHighlight = {
            x: (area.left - 1) * cellWidth + '%',
            y: (area.top - 1) * cellHeight + '%',
            height: area.height * cellHeight + '%',
            width: area.width * cellWidth + '%'
        };
        this.sanitizer.bypassSecurityTrustStyle(this.activeHighlight);
    }
    showGridLines(renderer) {
        if (renderer) {
            const cellHeight = renderer.grid.cellSize.height;
            const cellWidth = renderer.grid.cellSize.width;
            const height = cellHeight + '%';
            const width = cellWidth + '%';
            const rows = renderer.grid.rows;
            const cols = renderer.grid.columns;
            for (let i = 1; i < rows; i += 2) {
                let y, h, row;
                y = (i * cellHeight) + '%';
                h = 'calc(' + height + ' - 1px)';
                row = {
                    y: this.sanitizer.bypassSecurityTrustStyle(y),
                    height: this.sanitizer.bypassSecurityTrustStyle(h)
                };
                this.gridRows.push(row);
            }
            for (let i = 1; i < cols; i += 2) {
                let x, w, col;
                x = (i * cellWidth) + '%';
                w = 'calc(' + width + ' - 1px)';
                col = {
                    x: this.sanitizer.bypassSecurityTrustStyle(x),
                    width: this.sanitizer.bypassSecurityTrustStyle(w)
                };
                this.gridCols.push(col);
            }
        }
    }
}
NgxGridOverlayComponent.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "15.2.9", ngImport: i0, type: NgxGridOverlayComponent, deps: [{ token: i1.DomSanitizer }], target: i0.ɵɵFactoryTarget.Component });
NgxGridOverlayComponent.ɵcmp = i0.ɵɵngDeclareComponent({ minVersion: "14.0.0", version: "15.2.9", type: NgxGridOverlayComponent, selector: "ngx-grid-overlay", inputs: { renderer: "renderer", rows: "rows", cols: "cols", highlight: "highlight", showGrid: "showGrid" }, ngImport: i0, template: "<div class=\"wg-grid-overlay\">\n  <div *ngFor=\"let row of gridRows\"\n       class=\"wg-preview-item wg-preview-row\"\n       [style.height]=\"row.height\"\n       [style.top]=\"row.y\"></div>\n  <div *ngFor=\"let column of gridCols\"\n       class=\"wg-preview-item wg-preview-column\"\n       [style.width]=\"column.width\"\n       [style.left]=\"column.x\"></div>\n  <div *ngIf=\"activeHighlight\"\n       class=\"wg-preview-item wg-preview-highlight\"\n       [style.top]=\"activeHighlight.y\"\n       [style.height]=\"activeHighlight.height\"\n       [style.left]=\"activeHighlight.x\"\n       [style.width]=\"activeHighlight.width\"></div>\n</div>\n", styles: [".wg-grid-overlay{overflow:hidden}.wg-grid-overlay *{position:absolute;height:100%;width:100%}.wg-grid-overlay .wg-preview-item{position:absolute;display:inline-block}.wg-grid-overlay .wg-preview-item.wg-preview-row{width:100%;border-top:1px dotted rgb(215,215,215);border-bottom:1px solid rgb(240,240,240)}.wg-grid-overlay .wg-preview-item.wg-preview-column{height:100%;border-left:1px dotted rgb(215,215,215);border-right:1px solid rgb(240,240,240)}.wg-grid-overlay .wg-preview-item.wg-preview-highlight{background-color:#0071bc33;z-index:1}\n"], dependencies: [{ kind: "directive", type: i2.NgForOf, selector: "[ngFor][ngForOf]", inputs: ["ngForOf", "ngForTrackBy", "ngForTemplate"] }, { kind: "directive", type: i2.NgIf, selector: "[ngIf]", inputs: ["ngIf", "ngIfThen", "ngIfElse"] }], changeDetection: i0.ChangeDetectionStrategy.OnPush });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "15.2.9", ngImport: i0, type: NgxGridOverlayComponent, decorators: [{
            type: Component,
            args: [{ selector: 'ngx-grid-overlay', changeDetection: ChangeDetectionStrategy.OnPush, template: "<div class=\"wg-grid-overlay\">\n  <div *ngFor=\"let row of gridRows\"\n       class=\"wg-preview-item wg-preview-row\"\n       [style.height]=\"row.height\"\n       [style.top]=\"row.y\"></div>\n  <div *ngFor=\"let column of gridCols\"\n       class=\"wg-preview-item wg-preview-column\"\n       [style.width]=\"column.width\"\n       [style.left]=\"column.x\"></div>\n  <div *ngIf=\"activeHighlight\"\n       class=\"wg-preview-item wg-preview-highlight\"\n       [style.top]=\"activeHighlight.y\"\n       [style.height]=\"activeHighlight.height\"\n       [style.left]=\"activeHighlight.x\"\n       [style.width]=\"activeHighlight.width\"></div>\n</div>\n", styles: [".wg-grid-overlay{overflow:hidden}.wg-grid-overlay *{position:absolute;height:100%;width:100%}.wg-grid-overlay .wg-preview-item{position:absolute;display:inline-block}.wg-grid-overlay .wg-preview-item.wg-preview-row{width:100%;border-top:1px dotted rgb(215,215,215);border-bottom:1px solid rgb(240,240,240)}.wg-grid-overlay .wg-preview-item.wg-preview-column{height:100%;border-left:1px dotted rgb(215,215,215);border-right:1px solid rgb(240,240,240)}.wg-grid-overlay .wg-preview-item.wg-preview-highlight{background-color:#0071bc33;z-index:1}\n"] }]
        }], ctorParameters: function () { return [{ type: i1.DomSanitizer }]; }, propDecorators: { renderer: [{
                type: Input
            }], rows: [{
                type: Input
            }], cols: [{
                type: Input
            }], highlight: [{
                type: Input
            }], showGrid: [{
                type: Input
            }] } });

class NgxWidgetGridComponent {
    constructor(el, _changeDetector, _renderer) {
        this.el = el;
        this._changeDetector = _changeDetector;
        this._renderer = _renderer;
        this.showGrid = false;
        this.widgetPositionChange = new EventEmitter();
        this.gridFull = new EventEmitter();
        this.gridAlreadyFull = false;
        this._highlightNextPosition = false;
        this.grid = new Grid(this.rows, this.columns);
        this.gridRenderer = new GridRenderer(this.grid);
    }
    get rows() {
        return this._rows;
    }
    set rows(rows) {
        this._rows = rows;
        this.updateGridSize();
    }
    get columns() {
        return this._columns;
    }
    set columns(cols) {
        this._columns = cols;
        this.updateGridSize();
    }
    get highlightNextPosition() {
        return this._highlightNextPosition;
    }
    set highlightNextPosition(highlightNext) {
        this._highlightNextPosition = highlightNext;
        if (highlightNext) {
            this.updateNextPositionHighlight();
        }
        else {
            this.resetHighlights();
        }
    }
    ngAfterViewInit() {
        this.refreshWidgets();
        this.widgetComponents.changes.subscribe(() => {
            this.clearGrid();
            this.refreshWidgets();
        });
    }
    refreshWidgets() {
        this.widgetComponents.forEach((widget) => {
            if (!this.hasWidget(widget)) {
                this.addWidget(widget, true);
            }
            else {
            }
        });
        this.updateRendering();
    }
    hasWidget(widget) {
        return this.grid.hasWidget(widget.getConfig());
    }
    addWidget(widget, deferredRender) {
        this.grid.add(widget.getConfig());
        if (!deferredRender) {
            this.updateRendering();
        }
    }
    clearGrid() {
        this.grid.removeAll();
    }
    updateGridSize() {
        const columns = this.columns;
        const rows = this.rows;
        if (this.grid.columns !== columns || this.grid.rows !== rows) {
            this.grid.resize(rows, columns);
            this.updateRendering();
        }
    }
    updateRendering() {
        this.gridRenderer.render(this.grid, this.emitUpdatePosition.bind(this));
        this.updateNextPositionHighlight();
        // TODO: retrieve all widgets and call their updateRendering
        if (this.widgetComponents) {
            this.widgetComponents.forEach((widget) => {
                this.updateWidget(widget, false);
            });
        }
    }
    getGridRectangle() {
        const gridContainer = this.el.nativeElement;
        // c.f. jQuery#offset: https://github.com/jquery/jquery/blob/2d715940b9b6fdeed005cd006c8bf63951cf7fb2/src/offset.js#L93-105
        const rect = gridContainer.getBoundingClientRect();
        if (rect.width || rect.height || gridContainer.getClientRects().length) {
            const doc = gridContainer.ownerDocument;
            const docElem = doc.documentElement;
            return new Rectangle({
                top: rect.top + window.pageYOffset - docElem.clientTop,
                left: rect.left + window.pageXOffset - docElem.clientLeft,
                height: rect.height,
                width: rect.width
            });
        }
        return new Rectangle({ top: 0, left: 0, height: 0, width: 0 });
    }
    rasterizeCoords(x, y) {
        return this.gridRenderer.rasterizeCoords(x, y, this.el.nativeElement.clientWidth, this.el.nativeElement.clientHeight);
    }
    updateWidget(widget, swappingWidgets) {
        const config = widget.getConfig();
        const newPosition = config.position;
        const el = widget.getEl();
        this.gridRenderer.setWidgetPosition(config.id, newPosition, swappingWidgets);
        const widgetStyles = this.getWidgetStyle(widget);
        for (const style in widgetStyles) {
            if (widgetStyles.hasOwnProperty(style)) {
                this._renderer.setStyle(el.nativeElement, style, widgetStyles[style]);
            }
        }
        this.emitUpdatePosition(config);
        this.assessAvailableGridSpace();
    }
    getWidgetPositionByWidgetId(widgetId) {
        return this.gridRenderer.getWidgetPosition(widgetId);
    }
    getWidgetPosition(widget) {
        return this.getWidgetPositionByWidgetId(widget.getConfig().id);
    }
    getWidgetStyle(widget) {
        return this.gridRenderer.getStyle(widget.getConfig().id);
    }
    isPointObstructed(i, j) {
        return this.gridRenderer ? this.gridRenderer.isObstructed(i, j) : true;
    }
    isAreaObstructed(area, options) {
        return this.gridRenderer ? this.gridRenderer.isAreaObstructed(area, options) : true;
    }
    areaObstructor(area) {
        return this.gridRenderer.obstructions[(area.top - 1) * this.grid.columns + (area.left - 1)];
    }
    highlightArea(area) {
        if (area.top && area.left && area.height && area.width) {
            setTimeout(() => {
                this.highlightedArea = area;
                this._changeDetector.markForCheck();
            });
        }
    }
    updateNextPositionHighlight() {
        if (this.highlightNextPosition) {
            this.highlightedArea = this.gridRenderer.getNextPosition();
        }
    }
    getNextPosition() {
        return this.gridRenderer.getNextPosition();
    }
    getPositions() {
        return this.grid.widgets;
    }
    resetHighlights() {
        setTimeout(() => {
            this.highlightedArea = null;
            this._changeDetector.markForCheck();
        });
    }
    emitUpdatePosition(widget) {
        this.widgetPositionChange.emit({
            index: this.getWidgetIndex(widget),
            newPosition: widget.position
        });
    }
    getWidgetIndex(widgetConfig) {
        for (let i = this.grid.widgets.length - 1; i >= 0; i--) {
            if (this.grid.widgets[i].id === widgetConfig.id) {
                return i;
            }
        }
        return -1;
    }
    getWidgetById(widgetId) {
        return this.widgetComponents.find(widgetCmp => {
            return widgetCmp.getConfig().id === widgetId;
        }) || null;
    }
    assessAvailableGridSpace() {
        const gridHasSpaceLeft = this.gridRenderer.hasSpaceLeft();
        if (this.gridAlreadyFull) {
            if (gridHasSpaceLeft) {
                this.gridFull.emit(false);
                this.gridAlreadyFull = false;
            }
            else {
                /*No change to grid status. was and still is full*/
            }
        }
        else {
            if (!gridHasSpaceLeft) {
                this.gridFull.emit(true);
                this.gridAlreadyFull = true;
            }
            else {
                /*No change to grid status. had and still has available space*/
            }
        }
    }
}
NgxWidgetGridComponent.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "15.2.9", ngImport: i0, type: NgxWidgetGridComponent, deps: [{ token: i0.ElementRef }, { token: i0.ChangeDetectorRef }, { token: i0.Renderer2 }], target: i0.ɵɵFactoryTarget.Component });
NgxWidgetGridComponent.ɵcmp = i0.ɵɵngDeclareComponent({ minVersion: "14.0.0", version: "15.2.9", type: NgxWidgetGridComponent, selector: "ngx-widget-grid", inputs: { showGrid: "showGrid", rows: "rows", columns: "columns", highlightNextPosition: "highlightNextPosition" }, outputs: { widgetPositionChange: "widgetPositionChange", gridFull: "gridFull" }, queries: [{ propertyName: "widgetComponents", predicate: NgxWidgetComponent }], ngImport: i0, template: "<ngx-grid-overlay [showGrid]=\"showGrid\"\n                  [renderer]=\"gridRenderer\"\n                  [highlight]=\"highlightedArea\"\n                  [cols]=\"columns\"\n                  [rows]=\"rows\">\n</ngx-grid-overlay>\n<div class=\"wg-grid-widgets\">\n  <ng-content></ng-content>\n</div>\n", styles: [":host{height:100%;width:100%;position:absolute;overflow:hidden}\n"], dependencies: [{ kind: "component", type: NgxGridOverlayComponent, selector: "ngx-grid-overlay", inputs: ["renderer", "rows", "cols", "highlight", "showGrid"] }], changeDetection: i0.ChangeDetectionStrategy.OnPush });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "15.2.9", ngImport: i0, type: NgxWidgetGridComponent, decorators: [{
            type: Component,
            args: [{ selector: 'ngx-widget-grid', changeDetection: ChangeDetectionStrategy.OnPush, template: "<ngx-grid-overlay [showGrid]=\"showGrid\"\n                  [renderer]=\"gridRenderer\"\n                  [highlight]=\"highlightedArea\"\n                  [cols]=\"columns\"\n                  [rows]=\"rows\">\n</ngx-grid-overlay>\n<div class=\"wg-grid-widgets\">\n  <ng-content></ng-content>\n</div>\n", styles: [":host{height:100%;width:100%;position:absolute;overflow:hidden}\n"] }]
        }], ctorParameters: function () { return [{ type: i0.ElementRef }, { type: i0.ChangeDetectorRef }, { type: i0.Renderer2 }]; }, propDecorators: { showGrid: [{
                type: Input
            }], widgetPositionChange: [{
                type: Output
            }], gridFull: [{
                type: Output
            }], widgetComponents: [{
                type: ContentChildren,
                args: [NgxWidgetComponent]
            }], rows: [{
                type: Input
            }], columns: [{
                type: Input
            }], highlightNextPosition: [{
                type: Input
            }] } });

class NgxWidgetGridModule {
}
NgxWidgetGridModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "15.2.9", ngImport: i0, type: NgxWidgetGridModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
NgxWidgetGridModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "14.0.0", version: "15.2.9", ngImport: i0, type: NgxWidgetGridModule, declarations: [NgxGridOverlayComponent,
        NgxWidgetComponent,
        NgxWidgetGridComponent,
        NgxWidgetMoverDirective,
        NgxWidgetResizerDirective], imports: [CommonModule], exports: [NgxWidgetComponent,
        NgxWidgetGridComponent] });
NgxWidgetGridModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "15.2.9", ngImport: i0, type: NgxWidgetGridModule, imports: [CommonModule] });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "15.2.9", ngImport: i0, type: NgxWidgetGridModule, decorators: [{
            type: NgModule,
            args: [{
                    imports: [
                        CommonModule
                    ],
                    declarations: [
                        NgxGridOverlayComponent,
                        NgxWidgetComponent,
                        NgxWidgetGridComponent,
                        NgxWidgetMoverDirective,
                        NgxWidgetResizerDirective
                    ],
                    exports: [
                        NgxWidgetComponent,
                        NgxWidgetGridComponent
                    ]
                }]
        }] });

/*
 * Public API Surface of ngx-widget-grid
 */

/**
 * Generated bundle index. Do not edit.
 */

export { Cell, CellSize, Grid, GridRenderer, NgxGridOverlayComponent, NgxWidgetComponent, NgxWidgetGridComponent, NgxWidgetGridModule, PathIterator, Rectangle, WidgetConfig };
//# sourceMappingURL=ngx-widget-grid.mjs.map
